import java.io.*;

public class Maze {

    public class Room {
        private int xpos, ypos;
        private VisitedState state;
        private RoomList neighbors;

        Room(int x, int y){
            xpos = x;
            ypos = y;
            neighbors = new RoomList();
            state = VisitedState.UNSEEN;
        } 

	int getX() { return xpos; }
	int getY() { return ypos; }

        boolean Adjacent(Room otherRoom){
            RoomNode rn = neighbors.first;
            while (rn != null){
                if (rn.data == otherRoom) return true;
                rn = rn.next;
            }
            return false;
        }

        void setState(VisitedState s){
            state = s;
            // Hook to update a visualization would go right here
        }

	VisitedState getState() {
	    return state;
	}

        RoomList getNeighbors() { return neighbors; }

    }

    public class RoomNode {
        private Room data;
        private RoomNode next;
    
        RoomNode(Room r){
            data = r;
            next = null;
        }
    
	RoomNode(Room r, RoomNode n) {
	    data = r;
	    next = n;
	}

        Room getRoom(){ return data; }
        RoomNode getNext(){ return next; }
    }

    public class RoomList {
        private int length;
        private RoomNode first;

        RoomList(){
            length = 0;
            first = null;
        }

        void Insert(Room r){
            RoomNode rn = new RoomNode(r,first);
            first = rn;
            length++;
        }

        RoomNode getFirst(){ return first; }
    }

    // Beginning of class Maze proper
    private Room start, end;
    private int width, height;
    private Room [][] rooms;

    Room getRoomAt(int x, int y){
        return rooms[x][y];
    }

    Maze(String filename){
        // Create a new maze by reading from the named file
        EasyReader in = new EasyReader(filename);

        // Read the width and height
        width = in.readInt();
        height = in.readInt();
        in.readLine();

        // Create all the Room objects, store them in rooms
        rooms = new Room[width+1][height+1];
        for (int i=1; i<=width; i++)
            for (int j=1; j<=height; j++)
                rooms[i][j] = new Room(i,j);

        // Read the image of the maze and create its adjacency-list representation
        //    by populating the neighbors of each room and setting start and end.
        
        // Skip top line of walls
        in.readLine();
        
        for (int j=1; j<=height; j++){
            // Skip left wall
            in.readChar();
            // Process line of rooms, doors, and walls
          for (int i=1; i<=width; i++){
                char c = in.readChar(); // is a room
                if (c == '*') {
		    start = rooms[i][j];
		}
                else if (c == 'X') {
		    end = rooms[i][j];
		}
                c = in.readChar(); // next is wall or door
                if (i < width && c == ' '){
                    rooms[i][j].getNeighbors().Insert(rooms[i+1][j]);
                    rooms[i+1][j].getNeighbors().Insert(rooms[i][j]);
                }
            }
            // Now process line of walls and doors only
            in.readLine();
            in.readChar();
            for (int i=1; i<=width; i++){
                char c = in.readChar(); // is a wall or door
                if (j < height && c == ' '){
                    rooms[i][j].getNeighbors().Insert(rooms[i][j+1]);
                    rooms[i][j+1].getNeighbors().Insert(rooms[i][j]);
                }
                in.readChar(); // should always be a +
            }
            in.readLine();
        }
    }

    void printMaze(PrintStream out){
	// first, the top wall
	for (int i=1; i<=width; i++)
	    out.print("+-");
	out.println("+");

	for (int j=1; j<=height; j++) { 
	// line w/walls, rooms, and doors
	    out.print('|');
	    for (int i=1; i<=width; i++) {
		if (start == rooms[i][j])
		    out.print('*');
		else if (end == rooms[i][j])
		    out.print('X');
		else {
		    VisitedState v = rooms[i][j].getState();
		    if (v == VisitedState.UNSEEN)
			out.print(' ');
		    else if (v == VisitedState.VISITED)
			out.print('#');
		    else if (v == VisitedState.ONPATH)
			out.print('.');
		    else if (v == VisitedState.PROCESSING)
			out.print('?');
		}
		if (i < width) {
		    if (rooms[i][j].Adjacent(rooms[i+1][j]))
			out.print(' ');
		    else 
			out.print('|');
		}
	    }
	    out.println('|');

	    // line w/walls and doors only
	    for (int i=1; i<=width; i++) {
		if (i > 1 && j < height &&
		    rooms[i][j].Adjacent(rooms[i-1][j]) &&
		    rooms[i][j+1].Adjacent(rooms[i-1][j+1]) &&
		    rooms[i-1][j].Adjacent(rooms[i-1][j+1]) &&
		    rooms[i][j+1].Adjacent(rooms[i][j]))
		    out.print(' ');
		else
		    out.print('+');
		if (j < height && rooms[i][j].Adjacent(rooms[i][j+1]))
		    out.print(' ');
		else
		    out.print('-');
	    }
	    out.println('+');
	}
    }

}
